/*
 * Copyright (c) 2013, Altera Corporation.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  - Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

/* Global declarations shared by all files of this driver. */

#ifndef ACLPCI_H
#define ACLPCI_H


#include <linux/kobject.h>
#include <linux/kdev_t.h>
#include <linux/list.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/jiffies.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/uaccess.h>
#include <linux/sched.h>
#include <linux/version.h>

// https://lore.kernel.org/all/20220225181927.GA368545@bhelgaas/T/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,18,0)
#include <linux/dma-mapping.h>
#endif

/* includes from opencl/include/pcie */
#include "hw_pcie_constants.h"
#include "pcie_linux_driver_exports.h"

/* Local includes */
#include "version.h"

/* For now, disable DMA on big-endian system. */
#ifdef ACLPCI_BIGENDIAN
#define USE_DMA           0
#else
#define USE_DMA           1
#endif

#include "aclpci_dma.h"


#define DRIVER_NAME       "predator_driver"
#define BOARD_NAME        "predatorDev_"
#define DRIVER_VENDOR     "KAYA Instruments Ltd. / PCO"

// The following macros should only be applied to code that exists purely for debugging purposes
// and otherwise we want to suppress "not used" warnings
// (we don't want to apply global compiler switches because we do want to see those warnings where
// these macros are not explicitely applied

#ifdef UNUSED
#elif defined(__GNUC__)
# define UNUSED(x) UNUSED_ ## x __attribute__((unused))
#elif defined(__LCLINT__)
# define UNUSED(x) /*@unused@*/ x
#else
# define UNUSED(x) x
#endif


#ifdef __GNUC__
#define VARIABLE_IS_NOT_USED __attribute__ ((unused))
#else
#define VARIABLE_IS_NOT_USED
#endif

//---------------------------------------------------------------------------

/* Set to 1 to use Polling (instead of interrupts) to communicate
 * with hal. HAL must be compiled in the same mode */
#define POLLING           0


/* Debugging defines */
#define DO_DEBUG        0
#define VERBOSE_DEBUG   0
#if DO_DEBUG
#define ACL_DEBUG(...)					\
  do {						\
    printk("%s (%d): ", __func__, __LINE__);	\
    printk(__VA_ARGS__);			\
    printk("-----------------\n");				\
  } while (0)
#else
#define ACL_DEBUG(...)
#endif
#if VERBOSE_DEBUG
# define ACL_VERBOSE_DEBUG(...) ACL_DEBUG(__VA_ARGS__)
#else 
# define ACL_VERBOSE_DEBUG(...)
#endif

/* Don't actually bring down the kernel on an error condition */
#define assert(expr) \
do { \
   if (!(expr)) { \
      printk(KERN_ERR "Assertion failed! %s, %s, %s, line %d\n", \
            #expr, __FILE__, __func__, __LINE__); \
   } \
} while (0)



/* Maximum size of driver buffer (allocated with kalloc()).
 * Needed to copy data from user to kernel space, among other
 * things. */
static const size_t BUF_SIZE = PAGE_SIZE;


/* Device data used by this driver. */
struct aclpci_dev {
  /* the kernel pci device data structure */
  struct pci_dev *pci_dev;
  
  /* upstream root node */
  struct pci_dev *upstream;
  
  /* kernels virtual addr. for the mapped BARs */
  void * __iomem bar[ACL_PCI_NUM_BARS];
  
  /* length of each memory region. Used for error checking. */
  size_t bar_length[ACL_PCI_NUM_BARS];
  
  /* Controls which section of board's DDR maps to BAR */
  u64 global_mem_segment;
  
  /* Location of global_mem_segment value on the board. */
  void *global_mem_segment_addr;
  
  /* temporary buffer. If allocated, will be BUF_SIZE. */
  char *buffer;
  
  /* Mutex for this device. */
  struct semaphore sem;
  
  /* PID of process that called open() */
  int user_pid;
  
  /* character device */
  dev_t cdev_num;
  struct cdev cdev;
  struct class *my_class;
  struct device *device;


  /* signal sending structs */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,20,0)
  struct kernel_siginfo signal_info;
#else
  struct siginfo signal_info;
#endif
//old  struct task_struct *user_task;
//old  int user_filehandle;
  struct task_struct *user_task_for_irq[ACL_PCI_MAX_PROCESSES]; // assigned in init_irq() only, NOT in aclpci_open() !!!

  struct file *pFile[ACL_PCI_MAX_PROCESSES];
  struct task_struct *user_task_for_file[ACL_PCI_MAX_PROCESSES]; // assigned in aclpci_open()

  /* 1 if doing core reconfig via PCIe. 
   * Ignore all interrupts when this is going on. */
  int cvp_in_progress;
  
  /* State of uncorrectable error mask register, AER ext capability.
   * Saved during reprogramming */
  u32 aer_uerr_mask_reg;
  
  /* All the DMA data */
  //struct aclpci_dma dma_data;
  
  /* IRQ requests 'reference counter' */
  int irq_requested;

  /* Debug data */
  /* number of hw interrupts handled. */
  int num_handled_interrupts;
//old  int num_undelivered_signals;
  int num_undelivered_signals[ACL_PCI_MAX_PROCESSES];
  int pci_gen;
  int pci_num_lanes;
};


/* aclpci_fileio.c funciton */
int aclpci_open(struct inode *inode, struct file *pFile);
int aclpci_close(struct inode *inode, struct file *pFile);
ssize_t aclpci_read(struct file *file, char __user *buf, size_t count, loff_t *pos);
ssize_t aclpci_write(struct file *file, const char __user *buf, size_t count, loff_t *pos);
void* aclpci_get_checked_addr (int bar_id, void *device_addr, size_t count,
                               struct aclpci_dev *aclpci, ssize_t *errno, int print_error_msg);

/*loff_t aclpci_llseek(struct file *filp, loff_t off, int whence);*/

/* aclpci.c functions */
void load_signal_info (struct aclpci_dev *aclpci);
int init_irq (struct pci_dev *dev, void *dev_id);
//old int release_irq (struct pci_dev *dev, void *aclpci);
int release_irq (struct pci_dev *dev, void *dev_id, struct task_struct *user_task);

/* aclpci_dma.c functions */
void aclpci_dma_init(struct aclpci_dev *aclpci);
void aclpci_dma_finish(struct aclpci_dev *aclpci);
int aclpci_dma_get_idle_status(struct aclpci_dev *aclpci);
irqreturn_t aclpci_dma_service_interrupt (struct aclpci_dev *aclpci);
int aclpci_dma_update(struct aclpci_dev *aclpci, int forced);
int lock_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma);
void unlock_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma);
//void sync_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma);
void sync_cache_for_memory(struct aclpci_dev *aclpci, WD_DMA *dma, bool forDevice);

int cuda_lock_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma);
void cuda_unlock_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma);


/* aclpci_cmd.c functions */
void get_pcie_params (struct aclpci_dev *aclpci);
ssize_t aclpci_exec_cmd (struct aclpci_dev *aclpci, struct aclpci_cmd kcmd, size_t count);
int aclpci_get_user_pages(unsigned long start_page, size_t num_pages, struct page **p);
void aclpci_release_user_pages(struct page **p, size_t num_pages);

void get_wd_debug_info(WD_DEBUG_INFO* pWD_DEBUG_INFO);

#endif /* ACLPCI_H */
